@@ -16,6 +16,7 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <strings.h>
+#include <string.h>
 #include <unistd.h>
 #include <sys/stat.h>
 #if defined(USE_SYSLOG)
@@ -40,11 +41,12 @@
 #define TRUE 1
 #define FALSE 0
 
-extern int umountmain (int argc, char *argv[]);
+extern int umountmain(int argc, char *argv[]);
 
-int softdog;
+int softdog, softboot = FALSE;
 char *devname = DEVNAME;
 char *progname;
+int maxload1 = MAXLOAD, maxload5 = MAXLOAD * 3 / 4, maxload15 = MAXLOAD / 2;
 
 /* write a log entry on exit */
 void log_end()
@@ -85,9 +87,9 @@ void usage(void)
 {
     fprintf(stderr, "%s version %d.%d, usage:\n", progname, MAJOR_VERSION, MINOR_VERSION);
 #if defined(USE_SYSLOG)
-    fprintf(stderr, "%s [-i <interval> [-f]] [-n <file-name>] [-d <dev-name>] [-v] [-s]\n", progname);
+    fprintf(stderr, "%s [-i <interval> [-f]] [-n <file-name>] [-d <dev-name>] [-l <max loadavg>] [-v] [-s]\n", progname);
 #else				/* USE_SYSLOG */
-    fprintf(stderr, "%s [-i <interval> [-f]] [-n <file-name>] [-d <dev-name>] [-s]\n", progname);
+    fprintf(stderr, "%s [-i <interval> [-f]] [-n <file-name>] [-d <dev-name>] [-l <max loadavg>] [-s]\n", progname);
 #endif				/* USE_SYSLOG */
     exit(1);
 }
@@ -121,9 +123,9 @@ void mnt_off()
 
     fp = setmntent(MNTTAB, "r");
     while ((mnt = getmntent(fp)) != (struct mntent *) 0) {
-        /* First check if swap */
+	/* First check if swap */
 	if (!strcmp(mnt->mnt_type, MNTTYPE_SWAP))
-	    if (swapoff(mnt->mnt_fsname) < 0) 
+	    if (swapoff(mnt->mnt_fsname) < 0)
 		perror(mnt->mnt_fsname);
 
 	/* quota only if mounted at boot time && filesytem=ext2 */
@@ -132,12 +134,12 @@ void mnt_off()
 
 	/* group quota? */
 	if (hasmntopt(mnt, MNTOPT_GRPQUOTA))
-	    if (quotactl(QCMD(Q_QUOTAOFF, GRPQUOTA), mnt->mnt_fsname, 0, (caddr_t) 0) < 0) 
+	    if (quotactl(QCMD(Q_QUOTAOFF, GRPQUOTA), mnt->mnt_fsname, 0, (caddr_t) 0) < 0)
 		perror(mnt->mnt_fsname);
 
 	/* user quota */
 	if (hasmntopt(mnt, MNTOPT_USRQUOTA))
-	    if (quotactl(QCMD(Q_QUOTAOFF, USRQUOTA), mnt->mnt_fsname, 0, (caddr_t) 0) < 0) 
+	    if (quotactl(QCMD(Q_QUOTAOFF, USRQUOTA), mnt->mnt_fsname, 0, (caddr_t) 0) < 0)
 		perror(mnt->mnt_fsname);
 
     }
@@ -153,8 +155,18 @@ void myreboot(void)
 void do_shutdown(void)
 {
     int i = 0, fd;
-    char *uav[3]={"watchdog","-a",NULL};
-        
+    char *uav[3] =
+    {"watchdog", "-a", NULL};
+
+    /* soft-boot the system */
+    /* first close the watchdog device */
+    close_dev();
+#if defined(USE_SYSLOG)
+    /* now tell syslog what's happening */
+    syslog(LOG_INFO, "shutting down the system");
+    closelog();
+#endif				/* USE_SYSLOG */
+
     /* We cannot start shutdown, since init might not be able to fork. */
     /* That would stop the reboot process. So we try rebooting the system */
     /* ourselves. Note, that it is very likely we cannot start any rc */
@@ -178,7 +190,6 @@ void do_shutdown(void)
 	/* cause the reboot */
 	perror(progname);
     }
-
     /* Kill all processes. */
 
     (void) kill(-1, SIGTERM);
@@ -201,7 +212,6 @@ void do_shutdown(void)
 	write(fd, (char *) &wtmp, sizeof(wtmp));
 	close(fd);
     }
-
     /* Turn off accounting */
     if (acct(NULL) < 0)
 	perror(progname);
@@ -213,9 +223,9 @@ void do_shutdown(void)
     /* from the exit code. Changing the umount code to use returns would     */
     /* make much more work, in particular when upgrading the umount version. */
     if (!atexit(myreboot))
-    	umountmain(2,uav);
+	umountmain(2, uav);
     else
-    	myreboot();
+	myreboot();
 
     exit(0);
 }
@@ -227,19 +237,116 @@ void sync_system(void)
     sync();
 }
 
+/* check if process table is full */
+int check_fork(void)
+{
+    pid_t forkresult;
+    
+    forkresult = fork();
+    if (!forkresult)
+	exit(0);		/* child, exit immediately */
+    else if (forkresult < 0) {	/* fork failed */
+#if defined(USE_SYSLOG)
+	syslog(LOG_ERR, "process table is full!");
+#endif				/* USE_SYSLOG */
+	return (TRUE);
+    } else {
+	/* fork was okay          */
+	/* wait for child to stop */
+	if (waitpid(forkresult, NULL, 0) < 0) {
+#if defined(USE_SYSLOG)
+	    syslog(LOG_ERR, "child %d does not exist", forkresult);
+#else				/* USE_SYSLOG */
+	    perror(progname);
+#endif				/* USE_SYSLOG */
+	    if (softboot)
+		return (TRUE);
+	}
+    }
+    return (FALSE);
+}
+
+int check_file(char *filename)
+{
+    /* in filemode stat file */
+    if (filename != NULL) {
+	struct stat buf;
+
+	if (stat(filename, &buf) == -1) {
+#if defined(USE_SYSLOG)
+	    syslog(LOG_ERR, "cannot stat %s", filename);
+#else				/* USE_SYSLOG */
+	    perror(progname);
+#endif				/* USE_SYSLOG */
+	    if (softboot)
+		return (TRUE);
+	}
+    }
+    return (FALSE);
+}
+
+int check_load()
+{
+    int load, avg1, avg5, avg15;
+    char buf[40];
+
+    /* open the load average file */
+    load = open("/proc/loadavg", O_RDONLY);
+    if (load == -1) {
+#if defined(USE_SYSLOG)
+	syslog(LOG_ERR, "cannot open /proc/loadavg (errno =%d)", errno);
+#else				/* USE_SYSLOG */
+	perror(progname);
+#endif				/* USE_SYSLOG */
+	if (softboot)
+	    return (TRUE);
+    }
+    /* read the line (there is only one) */
+    if (read(load, buf, sizeof(buf)) < 0) {
+#if defined(USE_SYSLOG)
+	syslog(LOG_ERR, "read /proc/loadavg gave errno = %d", errno);
+#else				/* USE_SYSLOG */
+	perror(progname);
+#endif				/* USE_SYSLOG */
+	if (softboot)
+	    return (TRUE);
+    }
+    /* we only care about integer values */
+    avg1 = atoi(buf);
+    avg5 = atoi(strchr(buf, ' '));
+    avg15 = atoi(strchr(strchr(buf, ' ') + 1, ' '));
+
+    if (close(load) < 0) {
+#if defined(USE_SYSLOG)
+	syslog(LOG_ERR, "close /proc/loadavg gave errno = %d", errno);
+#else				/* USE_SYSLOG */
+	perror(progname);
+#endif				/* USE_SYSLOG */
+	if (softboot)
+	    return (TRUE);
+    }
+    if (avg1 > maxload1 || avg5 > maxload5 || avg15 > maxload15) {
+#if defined(USE_SYSLOG)
+	syslog(LOG_ERR, "loadavg %d %d %d > %d %d %d!", avg1, avg5, avg15,
+	       maxload1, maxload5, maxload15);
+#endif				/* USE_SYSLOG */
+	return (TRUE);
+    }
+    return (FALSE);
+}
+
 int main(int argc, char *const argv[])
 {
     FILE *fp;
     int c, tint = 10, force = FALSE, sync_it = FALSE;
-    int softboot = FALSE, boot_now = FALSE;
     char *filename = NULL;
     char log[128];
 #if defined(USE_SYSLOG)
-    char *opts = "d:i:n:fsvb";
+    char *opts = "d:i:n:fsvbl:";
     int verbose = FALSE;
     long count = 0L;
 #else				/* USE_SYSLOG */
-    char *opts = "d:i:n:fsb";
+    char *opts = "d:i:n:fsbl:";
 #endif				/* USE_SYSLOG */
 
     progname = basename(argv[0]);
@@ -264,6 +371,11 @@ int main(int argc, char *const argv[])
 	case 'b':
 	    softboot = TRUE;
 	    break;
+	case 'l':
+	    maxload1 = atoi(optarg);
+	    maxload5 = maxload1 * 3 / 4;
+	    maxload15 = maxload1 / 2;
+	    break;
 #if defined(USE_SYSLOG)
 	case 'v':
 	    verbose = TRUE;
@@ -273,26 +385,39 @@ int main(int argc, char *const argv[])
 	    usage();
 	}
     }
-
     if (tint < 0)
 	usage();
+
     if (tint >= TIMER_MARGIN && !force) {
 	fprintf(stderr, "%s warning:\n", progname);
 	fprintf(stderr, "This interval length might reboot the system while the process sleeps!\n");
 	fprintf(stderr, "To force this interval length use the -f option.\n");
 	exit(1);
     }
+    if (maxload1 < MINLOAD && !force) {
+	fprintf(stderr, "%s warning:\n", progname);
+	fprintf(stderr, "Using this maximal load average might reboot the system to often!\n");
+	fprintf(stderr, "To force this load average use the -f option.\n");
+	exit(1);
+    }
     if (fork())
 	exit(0);
 
     /* Okay, we're a daemon     */
+    /* but we're still attached to the tty */
+    setsid();
+
+    /* now we're free */
 #if defined(USE_SYSLOG)
+    /* with USE_SYSLOG we don't do any console IO */
+    close(0);
+    close(1);
+    close(2);
     /* Log the starting message */
     openlog(progname, LOG_PID, LOG_DAEMON);
     sprintf(log, "starting daemon (%d.%d): ", MAJOR_VERSION, MINOR_VERSION);
-    if (filename != NULL)
-	sprintf(log + strlen(log), "%s ", filename);
-    sprintf(log + strlen(log), "%ds%s%s", tint, sync_it ? " sync" : "", softboot ? " soft" : "");
+    sprintf(log + strlen(log), "%s%ds%s%s mla=%d", (filename == NULL) ? "" : filename,
+	    tint, sync_it ? " sync" : "", softboot ? " soft" : "", maxload1);
     syslog(LOG_INFO, log);
 #endif				/* USE_SYSLOG */
 
@@ -313,45 +438,21 @@ int main(int argc, char *const argv[])
 	fprintf(fp, "%d\n", getpid());
 	(void) fclose(fp);
     }
-    
     /* set signal term to call terminate() */
     /* to make sure softdog device is closed */
     signal(SIGTERM, terminate);
 
     /* main loop: update after <tint> seconds */
     while (1) {
-	pid_t forkresult;
-
-	/* write to the softdof device */
-	write(softdog, "\0", 1);
-
-	/* sync system if we have to */
-	if (sync_it)
-	    sync_system();
-
-	/* check if process table is full */
-	forkresult = fork();
-	if (!forkresult)
-	    exit(0);		/* child, exit immediately */
-	else if (forkresult < 0) {	/* fork failed */
+	/* write to the softdog device */
+	if (write(softdog, "\0", 1) < 0) {
 #if defined(USE_SYSLOG)
-	    syslog(LOG_ERR, "process table is full!");
-#endif				/* USE_SYSLOG */
-	    boot_now = TRUE;
-	} else {
-	    /* fork was okay          */
-	    /* wait for child to stop */
-	    if (waitpid(forkresult, NULL, 0) < 0) {
-#if defined(USE_SYSLOG)
-		syslog(LOG_ERR, "child %d does not exist", forkresult);
-#else				/* USE_SYSLOG */
-		perror(progname);
-#endif				/* USE_SYSLOG */
-		if (softboot)
-		    boot_now = TRUE;
-	    }
+	    syslog(LOG_ERR, "write watchdog device gave error %d!", errno);
+#endif
+	    if (softboot)
+		do_shutdown();
 	}
-	/* do verbose  logging */
+	/* do verbose logging */
 #if defined(USE_SYSLOG)
 	if (verbose) {
 	    count++;
@@ -359,33 +460,53 @@ int main(int argc, char *const argv[])
 	}
 #endif				/* USE _SYSLOG */
 
-	/* in filemode stat file */
-	if (filename != NULL) {
-	    struct stat buf;
+	/* sync system if we have to */
+	if (sync_it)
+	    sync_system();
 
-	    if (stat(filename, &buf) == -1) {
+	/* write to the softdog device */
+	if (write(softdog, "\0", 1) < 0) {
 #if defined(USE_SYSLOG)
-		syslog(LOG_ERR, "cannot stat %s", filename);
-#else				/* USE_SYSLOG */
-		perror(progname);
-#endif				/* USE_SYSLOG */
-		if (softboot)
-		    boot_now = TRUE;
-	    }
+	    syslog(LOG_ERR, "write watchdog device gave error %d!", errno);
+#endif
+	    if (softboot)
+		do_shutdown();
 	}
-	/* do we have to do a softboot? */
-	if (boot_now) {
-	    /* soft-boot the system */
-	    /* first close the device */
-	    close_dev();
-#if defined(USE_SYSLOG)
-	    /* now tell syslog what's happening */
-	    syslog(LOG_INFO, "shutting down the system");
-	    closelog();
-#endif				/* USE_SYSLOG */
-
-	    /* finally bring the system down */
+	/* check process table */
+	if (check_fork())
 	    do_shutdown();
+
+	/* write to the softdog device */
+	if (write(softdog, "\0", 1) < 0) {
+#if defined(USE_SYSLOG)
+	    syslog(LOG_ERR, "write watchdog device gave error %d!", errno);
+#endif
+	    if (softboot)
+		do_shutdown();
+	}
+	/* check load average */
+	if (check_load())
+	    do_shutdown();
+
+	/* write to the softdog device */
+	if (write(softdog, "\0", 1) < 0) {
+#if defined(USE_SYSLOG)
+	    syslog(LOG_ERR, "write watchdog device gave error %d!", errno);
+#endif
+	    if (softboot)
+		do_shutdown();
+	}
+	/* in filemode stat file */
+	if (check_file(filename))
+	    do_shutdown();
+
+	/* write to the softdog device */
+	if (write(softdog, "\0", 1) < 0) {
+#if defined(USE_SYSLOG)
+	    syslog(LOG_ERR, "write watchdog device gave error %d!", errno);
+#endif
+	    if (softboot)
+		do_shutdown();
 	}
 	/* finally sleep some seconds */
 	sleep(tint);
